@@ -29,10 +29,10 @@
 
 #include "ploop.h"
 
-int create_gpt_partition(const char *device, off_t size)
+int create_gpt_partition(const char *device, off_t size, __u32 blocksize)
 {
 	char cmd[512];
-	unsigned long long start = 1UL << PLOOP1_DEF_CLUSTER_LOG;
+	unsigned long long start = blocksize;
 	unsigned long long end = size - start;
 
 	if (size <= start) {
@@ -91,7 +91,7 @@ void tune_fs(const char *target, const char *device, unsigned long long size_sec
 	}
 	snprintf(cmd, sizeof(cmd), "/sbin/tune2fs -r %llu %s",
 			reserved_blocks, part_device);
-	ploop_log(0, "%s", cmd);
+	ploop_log(0, "Executing: %s", cmd);
 	system(cmd);
 }
 
@@ -128,7 +128,7 @@ int resize_fs(const char *device)
 			return ret;
 	}
 	snprintf(buf, sizeof(buf), "%s -p %s", prog, part_device);
-	ploop_err(0, "%s", buf);
+	ploop_err(0, "Executing: %s", buf);
 	ret = system(buf);
 	if (ret) {
 		ploop_err(0, "Failed to resize fs (cmd: %s)", buf);
