@@ -24,10 +24,12 @@
 #include <stdio.h>
 #include <time.h>
 
+#include "libploop.h"
+
 #define LOG_BUF_SIZE	8192
 static int _s_log_enable = 1;
 static int _s_log_level = 3;
-static int _s_log_verbose_level = -2; // disable stdout/stderr
+static int _s_log_verbose_level = PLOOP_LOG_NOCONSOLE; // disable stdout/stderr
 static FILE *_s_log_file = NULL;
 
 /* Thread Local Storage */
@@ -63,7 +65,8 @@ static void logger_ap(int level, int err_no, const char *format, va_list ap)
 			 strerror(err_no));
 	}
 	if (_s_log_enable) {
-		if (_s_log_verbose_level != -2 && _s_log_verbose_level >= level) {
+		if (_s_log_verbose_level != PLOOP_LOG_NOCONSOLE &&
+				_s_log_verbose_level >= level) {
 			fprintf(std, "%s\n", buf);
 			fflush(std);
 		}
@@ -116,15 +119,19 @@ void ploop_set_verbose_level(int level)
 
 int ploop_set_log_file(const char *fname)
 {
-	FILE *fp;
+	FILE *fp = NULL;
 
-	fp = fopen(fname, "a");
-	if (fp== NULL) {
-		ploop_err(errno, "Can't open %s", fname);
-		return -1;
+	if (fname != NULL) {
+		fp = fopen(fname, "a");
+		if (fp == NULL) {
+			ploop_err(errno, "Can't open %s", fname);
+			return -1;
+		}
 	}
+
 	if (_s_log_file != NULL)
 		fclose(_s_log_file);
 	_s_log_file = fp;
+
 	return 0;
 }
