@@ -46,6 +46,15 @@ struct ploop_cancel_handle *ploop_get_cancel_handle(void)
 	return &__cancel_data;
 }
 
+static int not_supported_for_vm(struct ploop_disk_images_data *di)
+{
+	if (!di->runtime->vm_compat)
+		return 0;
+
+	ploop_err(0, "Operation is not supported for Vm");
+	return SYSEXIT_PARAM;
+}
+
 /* set cancel flag
  * Note: this function also clear the flag
  */
@@ -229,23 +238,26 @@ static int get_temp_mountpoint(const char *file, int create, char *buf, int len)
 	return 0;
 }
 
-static int create_empty_delta(const char *path, off_t bdsize)
+static int create_empty_delta(const char *path, __u32 blocksize, off_t bdsize)
 {
 	int fd;
 	void * buf = NULL;
 	struct ploop_pvd_header *vh;
 	__u32 SizeToFill;
+	__u64 cluster = S2B(blocksize);
+
+	assert(blocksize);
 
 	if (bdsize > (__u32)-1)
 		return -1;
 
-	if (posix_memalign(&buf, 4096, CLUSTER)) {
+	if (posix_memalign(&buf, 4096, cluster)) {
 		ploop_err(errno, "posix_memalign");
 		return -1;
 	}
 
-	ploop_log(0, "Creating delta %s size=%ld sectors",
-			path, (long)bdsize);
+	ploop_log(0, "Creating delta %s bs=%d size=%ld sectors",
+			path, blocksize, (long)bdsize);
 	fd = open(path, O_RDWR|O_CREAT|O_EXCL, 0600);
 	if (fd < 0) {
 		ploop_err(errno, "Can't open %s", path);
@@ -253,25 +265,25 @@ static int create_empty_delta(const char *path, off_t bdsize)
 		return -1;
 	}
 
-	memset(buf, 0, CLUSTER);
+	memset(buf, 0, cluster);
 
 	vh = buf;
-	SizeToFill = generate_pvd_header(vh, bdsize);
+	SizeToFill = generate_pvd_header(vh, bdsize, blocksize);
 	vh->m_Flags = CIF_Empty;
 
-	if (WRITE(fd, buf, CLUSTER))
+	if (WRITE(fd, buf, cluster))
 		goto out_close;
 
-	if (SizeToFill > CLUSTER) {
+	if (SizeToFill > cluster) {
 		int i;
-		memset(buf, 0, CLUSTER);
-		for (i = 1; i < SizeToFill / CLUSTER; i++)
-			if (WRITE(fd, buf, CLUSTER))
+		memset(buf, 0, cluster);
+		for (i = 1; i < SizeToFill / cluster; i++)
+			if (WRITE(fd, buf, cluster))
 				goto out_close;
 	}
 
 	if (fsync(fd)) {
-		ploop_err(errno, "fsync");
+		ploop_err(errno, "fsync %s", path);
 		goto out_close;
 	}
 	free(buf);
@@ -285,7 +297,7 @@ out_close:
 	return -1;
 }
 
-static int create_empty_preallocated_delta(const char *path, off_t bdsize)
+static int create_empty_preallocated_delta(const char *path, __u32 blocksize, off_t bdsize)
 {
 	struct delta odelta = {};
 	int rc, clu, i;
@@ -294,57 +306,48 @@ static int create_empty_preallocated_delta(const char *path, off_t bdsize)
 	__u32 SizeToFill;
 	__u32 l2_slot = 0;
 	off_t off;
+	__u64 cluster = S2B(blocksize);
 
 	if (bdsize > (__u32)-1)
 		return -1;
 
-	if (posix_memalign(&buf, 4096, CLUSTER)) {
+	if (posix_memalign(&buf, 4096, cluster)) {
 		ploop_err(errno, "posix_memalign");
 		return -1;
 	}
 
-	ploop_log(0, "Creating preallocated delta %s size=%ld sectors",
-			path, (long)bdsize);
+	ploop_log(0, "Creating preallocated delta %s bs=%d size=%ld sectors",
+			path, blocksize, (long)bdsize);
 	rc = open_delta_simple(&odelta, path, O_RDWR|O_CREAT|O_EXCL, OD_OFFLINE);
 	if (rc) {
 		free(buf);
 		return -1;
 	}
 
-	memset(buf, 0, CLUSTER);
-	SizeToFill = generate_pvd_header(&vh, bdsize);
+	memset(buf, 0, cluster);
+	SizeToFill = generate_pvd_header(&vh, bdsize, blocksize);
 	memcpy(buf, &vh, sizeof(struct ploop_pvd_header));
 	vh.m_Flags = CIF_Empty;
 
-	if (sys_fallocate(odelta.fd, 0, 0, (off_t)(vh.m_FirstBlockOffset + vh.m_SizeInSectors) << 9)) {
+	if (sys_fallocate(odelta.fd, 0, 0, S2B(vh.m_FirstBlockOffset + vh.m_SizeInSectors))) {
 		ploop_err(errno, "Failed to fallocate");
 		goto out_close;
 	}
 
-	for (clu = 0; clu < SizeToFill / CLUSTER; clu++) {
+	for (clu = 0; clu < SizeToFill / cluster; clu++) {
 		if (is_operation_cancelled())
 			goto out_close;
 
 		if (clu > 0)
-			memset(buf, 0, CLUSTER);
-		for (i = (clu == 0 ? 16 : 0); i < (CLUSTER / sizeof(__u32)) &&
+			memset(buf, 0, cluster);
+		for (i = (clu == 0 ? 16 : 0); i < (cluster / sizeof(__u32)) &&
 				l2_slot < vh.m_Size;
 				i++, l2_slot++)
 		{
-			off = vh.m_FirstBlockOffset + (l2_slot * (1<<PLOOP1_DEF_CLUSTER_LOG));
+			off = vh.m_FirstBlockOffset + (l2_slot * blocksize);
 			((__u32*)buf)[i] = off;
 		}
-		if (WRITE(odelta.fd, buf, CLUSTER))
-			goto out_close;
-	}
-	// FIXME: workaround until ploop will support fallocate
-	memset(buf, 0, CLUSTER);
-	for (i = 0; i< vh.m_Size; i++) {
-		if (is_operation_cancelled())
-			goto out_close;
-
-		off_t off = (off_t)vh.m_FirstBlockOffset + (i * 1<<PLOOP1_DEF_CLUSTER_LOG);
-		if (PWRITE(&odelta, buf, CLUSTER, off << 9))
+		if (WRITE(odelta.fd, buf, cluster))
 			goto out_close;
 	}
 
@@ -372,7 +375,7 @@ static int create_raw_delta(const char * path, off_t bdsize)
 	ploop_log(0, "Creating raw delta %s size=%ld sectors",
 			path, (long)bdsize);
 
-	if (posix_memalign(&buf, 4096, CLUSTER)) {
+	if (posix_memalign(&buf, 4096, DEF_CLUSTER)) {
 		ploop_err(errno, "posix_memalign");
 		return -1;
 	}
@@ -384,15 +387,15 @@ static int create_raw_delta(const char * path, off_t bdsize)
 		return -1;
 	}
 
-	memset(buf, 0, CLUSTER);
+	memset(buf, 0, DEF_CLUSTER);
 
 	pos = 0;
 	while (pos < bdsize) {
 		if (is_operation_cancelled())
 			goto out_close;
 		off_t copy = bdsize - pos;
-		if (copy > CLUSTER/512)
-			copy = CLUSTER/512;
+		if (copy > DEF_CLUSTER/512)
+			copy = DEF_CLUSTER/512;
 		if (WRITE(fd, buf, copy*512))
 			goto out_close;
 		pos += copy;
@@ -439,7 +442,7 @@ static void fill_diskdescriptor(struct ploop_pvd_header *vh, struct ploop_disk_i
 }
 
 static int create_image(struct ploop_disk_images_data *di,
-		const char *file, off_t size_sec, int mode)
+		const char *file, __u32 blocksize, off_t size_sec, int mode)
 {
 	int fd = -1;
 	int ret;
@@ -462,14 +465,14 @@ static int create_image(struct ploop_disk_images_data *di,
 	if (mode == PLOOP_RAW_MODE)
 		fd = create_raw_delta(file, size_sec);
 	else if (mode == PLOOP_EXPANDED_MODE)
-		fd = create_empty_delta(file, size_sec);
+		fd = create_empty_delta(file, blocksize, size_sec);
 	else if (mode == PLOOP_EXPANDED_PREALLOCATED_MODE)
-		fd = create_empty_preallocated_delta(file, size_sec);
+		fd = create_empty_preallocated_delta(file, blocksize, size_sec);
 	if (fd < 0)
 		goto err;
 	close(fd);
 
-	generate_pvd_header(&vh, size_sec);
+	generate_pvd_header(&vh, size_sec, blocksize);
 	fill_diskdescriptor(&vh, di);
 
 	if (realpath(file, fname) == NULL) {
@@ -547,7 +550,7 @@ static int ploop_init_image(struct ploop_disk_images_data *di, struct ploop_crea
 		if (ret)
 			goto err;
 
-		ret = create_gpt_partition(mount_param.device, size);
+		ret = create_gpt_partition(mount_param.device, size, di->blocksize);
 		if (ret)
 			goto err;
 	}
@@ -559,14 +562,9 @@ static int ploop_init_image(struct ploop_disk_images_data *di, struct ploop_crea
 		goto err;
 
 err:
-	if (ploop_umount(mount_param.device, NULL)) {
+	if (ploop_umount_image(di))
 		if (ret == 0)
 			ret = SYSEXIT_UMOUNT;
-	} else {
-		char dev[MAXPATHLEN];
-		ploop_get_base_delta_uuid_fname(di->images[0]->file, dev, sizeof(dev));
-		unlink(dev);
-	}
 
 	return ret;
 }
@@ -600,12 +598,22 @@ int ploop_create_image(struct ploop_create_param *param)
 {
 	struct ploop_disk_images_data *di;
 	int ret;
+	__u32 blocksize;
+
+	blocksize = param->blocksize ?
+			param->blocksize : (1 << PLOOP1_DEF_CLUSTER_LOG);
+	if (!is_valid_blocksize(blocksize)) {
+		ploop_err(0, "Incorrect blocksize specified: %d",
+				blocksize);
+		return SYSEXIT_PARAM;
+	}
 
 	di = ploop_alloc_diskdescriptor();
 	if (di == NULL)
 		return SYSEXIT_NOMEM;
-
-	ret = create_image(di, param->image, param->size, param->mode);
+	di->blocksize = blocksize;
+	ret = create_image(di, param->image, di->blocksize,
+			param->size, param->mode);
 	if (ret)
 		return ret;
 	if (param->fstype != NULL) {
@@ -623,24 +631,23 @@ int ploop_getdevice(int *minor)
 	int lfd;
 	struct stat st;
 	struct ploop_getdevice_ctl req;
-	int err, ret = 0;
+	int ret = 0;
+	const char *dev = "/dev/ploop0";
 
+	if (stat(dev, &st))
+		if (mknod(dev, S_IFBLK, gnu_dev_makedev(PLOOP_DEV_MAJOR, 0))) {
+			ploop_err(errno, "mknod %s", dev);
+			return SYSEXIT_MKNOD;
+		}
 
-	if (stat("/dev/ploop0", &st))
-		mknod("/dev/ploop0", S_IFBLK, gnu_dev_makedev(PLOOP_DEV_MAJOR, 0));
-
-	lfd = open("/dev/ploop0", O_RDONLY);
+	lfd = open(dev, O_RDONLY);
 	if (lfd < 0) {
-		err = errno;
-		ploop_err(errno, "Can't open device /dev/ploop0");
-		errno = err;
+		ploop_err(errno, "Can't open device %s", dev);
 		return SYSEXIT_DEVICE;
 	}
 
 	if (ioctl(lfd, PLOOP_IOC_GETDEVICE, &req) < 0) {
-		err = errno;
 		ploop_err(errno, "PLOOP_IOC_GETDDEVICE");
-		errno = err;
 		ret = SYSEXIT_DEVIOC;
 	}
 	*minor = req.minor;
@@ -685,12 +692,10 @@ int delete_deltas(int devfd, const char *devname)
 		return errno;
 
 	while (top >= 0) {
-		int err = ioctl(devfd, PLOOP_IOC_DEL_DELTA, &top);
-		if (err < 0) {
-			err = errno;
+		if (ioctl(devfd, PLOOP_IOC_DEL_DELTA, &top) < 0) {
 			ploop_err(errno, "PLOOP_IOC_DEL_DELTA dev=%s lvl=%d",
 					devname, top);
-			return err;
+			return errno;
 		}
 		top--;
 	}
@@ -812,20 +817,6 @@ int ploop_get_dev_by_mnt(const char *path, char *buf, int size)
 	return get_dev_by_mnt(path, 1, buf, size);
 }
 
-void ploop_resolve_magic_fname(char *fname)
-{
-	char *p;
-
-	p = strrchr(fname, '.');
-	if (p != NULL && strcmp(p + 1, BASE_UUID_MAGIC) == 0)
-		*p = 0;
-}
-
-void ploop_get_base_delta_uuid_fname(char *delta, char *out, int len)
-{
-	snprintf(out, len, "%s."BASE_UUID_MAGIC, delta);
-}
-
 char *ploop_get_base_delta_uuid(struct ploop_disk_images_data *di)
 {
 	int i;
@@ -852,37 +843,15 @@ int ploop_get_top_delta_fname(struct ploop_disk_images_data *di, char *out, int
 		return -1;
 	}
 	if (snprintf(out, len, "%s", fname) > len -1) {
-		ploop_err(0, "not enough space to store data");
+		ploop_err(0, "Not enough space to store data");
 		return -1;
 	}
 	return 0;
 }
 
-/* Image can be mounted readonly several times and new ploop device is used
- * To assocoate ploop device with Diskdescriptor.xml we create base delta magic file,
- * For any action on Diskdescriptor.xml we will find the device by this file
- */
-static int create_base_magic_delta(char **delta)
-{
-	char fname[MAXPATHLEN];
-
-	ploop_get_base_delta_uuid_fname(*delta, fname, sizeof(fname));
-	unlink(fname);
-	if (link(*delta, fname)) {
-		ploop_err(errno, "Failed to create link %s %s",
-				*delta, fname);
-		return SYSEXIT_CREAT;
-	}
-	free(*delta);
-	*delta = strdup(fname);
-
-	return 0;
-}
-
 int ploop_find_dev_by_uuid(struct ploop_disk_images_data *di,
 		int check_state, char *out, int len)
 {
-	char fname[MAXPATHLEN];
 	int ret;
 	int running = 0;
 
@@ -890,8 +859,8 @@ int ploop_find_dev_by_uuid(struct ploop_disk_images_data *di,
 		ploop_err(0, "No images found in " DISKDESCRIPTOR_XML);
 		return -1;
 	}
-	ploop_get_base_delta_uuid_fname(di->images[0]->file, fname, sizeof(fname));
-	ret = ploop_find_dev_by_delta(fname, out, len);
+	ret = ploop_find_dev(di->runtime->component_name, di->images[0]->file,
+			out, len);
 	if (ret == 0 && check_state) {
 		if (ploop_get_attr(out, "running", &running)) {
 			ploop_err(0, "Can't get running attr for %s",
@@ -958,7 +927,7 @@ static int ploop_mount_fs(struct ploop_mount_param *param)
 			param->quota ? "usrjquota=aquota.user,grpjquota=aquota.group,jqfmt=vfsv0," : "",
 			param->mount_data ? param->mount_data : "");
 
-	ploop_log(0, "mount %s at %s fstype=%s data='%s' %s",
+	ploop_log(0, "Mounting %s at %s fstype=%s data='%s' %s",
 			part_device, param->target, fstype,
 			data, param->ro  ? "ro":"");
 
@@ -995,19 +964,47 @@ static int add_delta(int lfd, char *image, struct ploop_ctl_delta *req)
 	return 0;
 }
 
+static int create_ploop_dev(int minor)
+{
+	char device[64];
+	char devicep1[64];
+	struct stat st;
+
+	strcpy(device, "/dev/");
+	make_sysfs_dev_name(minor, device + 5, sizeof(device) - 5);
+	/* Create pair /dev/ploopN & /dev/ploopNp1 */
+	if (stat(device, &st)) {
+		if (mknod(device, S_IFBLK, gnu_dev_makedev(PLOOP_DEV_MAJOR, minor))) {
+			ploop_err(errno, "mknod %s", device);
+			return SYSEXIT_MKNOD;
+		}
+		chmod(device, 0600);
+	}
+	snprintf(devicep1, sizeof(devicep1), "%sp1", device);
+	if (stat(devicep1, &st)) {
+		if (mknod(devicep1, S_IFBLK, gnu_dev_makedev(PLOOP_DEV_MAJOR, minor+1))) {
+			ploop_err(errno, "mknod %s", devicep1);
+			return SYSEXIT_MKNOD;
+		}
+		chmod(devicep1, 0600);
+	}
+	return 0;
+}
+
 /* NB: caller will take care about *lfd_p even if we fail */
-static int add_deltas(char **images, struct ploop_mount_param *param, int raw,
-			    int *lfd_p)
+static int add_deltas(struct ploop_disk_images_data *di,
+		char **images, struct ploop_mount_param *param,
+		int raw, __u32 blocksize, int *lfd_p)
 {
 	int lckfd = -1;
 	char *device = param->device;
 	int i;
 	int ret = 0;
+	int registered = 0;
 
 	if (device[0] == '\0') {
 		char buf[64];
 		int minor;
-		struct stat st;
 
 		lckfd = ploop_global_lock();
 		if (lckfd == -1)
@@ -1018,25 +1015,24 @@ static int add_deltas(char **images, struct ploop_mount_param *param, int raw,
 			goto err;
 		snprintf(device, sizeof(param->device), "/dev/%s",
 				make_sysfs_dev_name(minor, buf, sizeof(buf)));
-		if (stat(device, &st)) {
-			char devicep1[64];
-
-			/* Create pair /dev/ploopN & /dev/ploopNp1 */
-			if (mknod(device, S_IFBLK, gnu_dev_makedev(PLOOP_DEV_MAJOR, minor)))
-				ploop_err(errno, "mknod %s", device);
-			chmod(device, 0600);
-			snprintf(devicep1, sizeof(devicep1), "%sp1", device);
-			if (mknod(devicep1, S_IFBLK, gnu_dev_makedev(PLOOP_DEV_MAJOR, minor+1)))
-				ploop_err(errno, "mknod %s", devicep1);
-			chmod(devicep1, 0600);
+		ret = create_ploop_dev(minor);
+		if (ret)
+			goto err;
+		if (di != NULL) {
+			ret = register_ploop_dev(di->runtime->component_name,
+					images[0], device);
+			if (ret)
+				goto err;
+			registered = 1;
 		}
+
 	}
 
 	*lfd_p = open(device, O_RDONLY);
 	if (*lfd_p < 0) {
 		ploop_err(errno, "Can't open device %s", device);
 		ret = SYSEXIT_DEVICE;
-		goto err;
+		goto err1;
 	}
 
 	for (i = 0; images[i] != NULL; i++) {
@@ -1051,7 +1047,7 @@ static int add_deltas(char **images, struct ploop_mount_param *param, int raw,
 			req.c.pctl_format = PLOOP_FMT_RAW;
 		if (ro)
 			req.c.pctl_flags = PLOOP_FMT_RDONLY;
-		req.c.pctl_cluster_log = 9;
+		req.c.pctl_cluster_log = ffs(blocksize) - 1;
 		req.c.pctl_size = 0;
 		req.c.pctl_chunks = 1;
 
@@ -1065,7 +1061,7 @@ static int add_deltas(char **images, struct ploop_mount_param *param, int raw,
 		ret = add_delta(*lfd_p, image, &req);
 		if (ret) {
 			i--; /* image is not added */
-			goto err;
+			goto err1;
 		}
 	}
 	if (ioctl(*lfd_p, PLOOP_IOC_START, 0) < 0) {
@@ -1075,7 +1071,7 @@ static int add_deltas(char **images, struct ploop_mount_param *param, int raw,
 		goto err;
 	}
 
-err:
+err1:
 	if (ret && *lfd_p != -1) {
 		int err = 0;
 
@@ -1089,16 +1085,22 @@ err:
 		if (err == 0 && ioctl(*lfd_p, PLOOP_IOC_CLEAR, 0) < 0)
 			ploop_err(errno, "PLOOP_IOC_CLEAR");
 	}
+	if (ret && registered)
+		unregister_ploop_dev(di->runtime->component_name, images[0]);
+err:
 	ploop_unlock(&lckfd);
 	return ret;
 }
 
-int ploop_mount(char **images, struct ploop_mount_param *param, int raw)
+int ploop_mount(struct ploop_disk_images_data *di, char **images,
+		struct ploop_mount_param *param, int raw)
 {
 	int lfd = -1;
 	struct stat st;
 	int i;
 	int ret = 0;
+	__u32 blocksize = 0;
+	__u32 prev_blocksize = 0;
 
 	if (images == NULL || images[0] == NULL) {
 		ploop_err(0, "ploop_mount: no deltas to mount");
@@ -1117,15 +1119,30 @@ int ploop_mount(char **images, struct ploop_mount_param *param, int raw)
 			continue;
 
 		ro  = (images[i+1] != NULL || param->ro) ? 1 : 0;
-		ret = ploop_fsck(images[i], 0, 0, 1, ro, 0);
+		ret = ploop_fsck(images[i], 0, 0, 1, ro, 0, &blocksize);
 		if (ret) {
 			ploop_err(0, "%s (%s): irrecoverable errors",
 					images[i], ro ? "ro" : "rw");
 			goto err;
 		}
+		if (raw && i == 1)
+			prev_blocksize = blocksize;
+		if (i != 0 && prev_blocksize != blocksize) {
+			ploop_err(0, "Incorrect blocksize %s bs=%d [prev bs=%d]",
+					images[i], blocksize, prev_blocksize);
+			ret = SYSEXIT_PARAM;
+			goto err;
+		}
+		prev_blocksize = blocksize;
+	}
+	if (di != NULL && (raw && i != 1) && di->blocksize != blocksize) {
+		 ploop_err(0, "Inconsistent di blocksize=%d != image blocksize=%d",
+				 di->blocksize, blocksize);
+		 ret = SYSEXIT_PARAM;
+		 goto err;
 	}
 
-	ret = add_deltas(images, param, raw, &lfd);
+	ret = add_deltas(di, images, param, raw, blocksize, &lfd);
 
 	if (ret)
 		goto err;
@@ -1171,17 +1188,7 @@ static int mount_image(struct ploop_disk_images_data *di, struct ploop_mount_par
 	images = make_images_list(di, guid, 0);
 	if (images == NULL)
 		return SYSEXIT_NOMEM;
-	if (!(flags & PLOOP_MOUNT_SNAPSHOT)) {
-		ret = create_base_magic_delta(&images[0]);
-		if (ret)
-			goto err;
-	}
-
-	ret = ploop_mount(images, param, di->raw);
-	if (ret)
-		unlink(images[0]);
-
-err:
+	ret = ploop_mount(di, images, param, di->raw);
 	free_images_list(images);
 
 	return ret;
@@ -1289,6 +1296,9 @@ int ploop_umount(const char *device, struct ploop_disk_images_data *di)
 	ret = ploop_stop(lfd, device);
 	close(lfd);
 
+	if (ret == 0 && di != NULL)
+		unregister_ploop_dev(di->runtime->component_name, di->images[0]->file);
+
 	return ret;
 }
 
@@ -1317,17 +1327,13 @@ int ploop_umount_image(struct ploop_disk_images_data *di)
 	}
 
 	ret = ploop_umount(dev, di);
-	if (ret == 0) {
-		ploop_get_base_delta_uuid_fname(di->images[0]->file, dev, sizeof(dev));
-		unlink(dev);
-	}
 
 	ploop_unlock_di(di);
 
 	return ret;
 }
 
-int ploop_grow_device(const char *device, off_t new_size)
+int ploop_grow_device(const char *device, __u32 blocksize, off_t new_size)
 {
 	int fd, ret;
 	struct ploop_ctl ctl;
@@ -1356,7 +1362,7 @@ int ploop_grow_device(const char *device, off_t new_size)
 
 	memset(&ctl, 0, sizeof(ctl));
 
-	ctl.pctl_cluster_log = 9;
+	ctl.pctl_cluster_log = ffs(blocksize) - 1;
 	ctl.pctl_size = new_size;
 
 	if (ioctl(fd, PLOOP_IOC_GROW, &ctl) < 0) {
@@ -1429,8 +1435,8 @@ int ploop_resize_image(struct ploop_disk_images_data *di, struct ploop_resize_pa
 			goto err;
 		}
 
-		new_balloon_size = balloon_size + (fs.f_bfree * fs.f_bsize >> 9);
-		new_balloon_size -= delta >> 9;
+		new_balloon_size = balloon_size + B2S(fs.f_bfree * fs.f_bsize);
+		new_balloon_size -= B2S(delta);
 		ret = ploop_balloon_change_size(mount_param.device,
 				balloonfd, new_balloon_size);
 	} else if (param->size > dev_size) {
@@ -1441,7 +1447,7 @@ int ploop_resize_image(struct ploop_disk_images_data *di, struct ploop_resize_pa
 			if (ret)
 				goto err;
 		}
-		ret = ploop_grow_device(mount_param.device, param->size);
+		ret = ploop_grow_device(mount_param.device, di->blocksize, param->size);
 		if (ret)
 			goto err;
 		ret = resize_fs(mount_param.device);
@@ -1465,7 +1471,7 @@ int ploop_resize_image(struct ploop_disk_images_data *di, struct ploop_resize_pa
 err:
 	close(balloonfd);
 	if (were_mounted)
-		ploop_umount(mount_param.device, NULL);
+		ploop_umount(mount_param.device, di);
 	ploop_unlock_di(di);
 	free_mount_param(&mount_param);
 
@@ -1476,16 +1482,24 @@ static int expanded2raw(struct ploop_disk_images_data *di)
 {
 	struct delta delta = {};
 	struct delta odelta = {};
-	int off = 16;
 	__u32 clu;
-	char buf[CLUSTER];
+	void *buf = NULL;
 	char tmp[MAXPATHLEN];
 	char fname[MAXPATHLEN];
 	int ret = -1;
+	__u64 cluster;
 
 	ploop_log(0, "Converting image to raw...");
+	// FIXME: deny snapshots
 	if (open_delta(&delta, di->images[0]->file, O_RDONLY, OD_OFFLINE))
 		return SYSEXIT_OPEN;
+	cluster = S2B(delta.blocksize);
+
+	if (posix_memalign(&buf, 4096, cluster)) {
+		ploop_err(errno, "posix_memalign");
+		goto err;
+	}
+
 	snprintf(tmp, sizeof(tmp), "%s.tmp",
 			di->images[0]->file);
 	ret = open_delta_simple(&odelta, tmp, O_RDWR|O_CREAT|O_EXCL|O_TRUNC, OD_OFFLINE);
@@ -1493,8 +1507,8 @@ static int expanded2raw(struct ploop_disk_images_data *di)
 		goto err;
 
 	for (clu = 0; clu < delta.l2_size; clu++) {
-		int l2_cluster = (clu + off) / (CLUSTER / sizeof(__u32));
-		__u32 l2_slot  = (clu + off) % (CLUSTER / sizeof(__u32));
+		int l2_cluster = (clu + PLOOP_MAP_OFFSET) / (cluster / sizeof(__u32));
+		__u32 l2_slot  = (clu + PLOOP_MAP_OFFSET) % (cluster / sizeof(__u32));
 
 		if (l2_cluster >= delta.l1_size) {
 			ploop_err(0, "abort: l2_cluster >= delta.l1_size");
@@ -1503,23 +1517,23 @@ static int expanded2raw(struct ploop_disk_images_data *di)
 		}
 
 		if (delta.l2_cache != l2_cluster) {
-			if (PREAD(&delta, delta.l2, CLUSTER, (off_t)l2_cluster * CLUSTER))
+			if (PREAD(&delta, delta.l2, cluster, (off_t)l2_cluster * cluster))
 				goto err;
 			delta.l2_cache = l2_cluster;
 		}
-		if ((delta.l2[l2_slot] % (1 << PLOOP1_DEF_CLUSTER_LOG)) != 0) {
+		if ((delta.l2[l2_slot] % delta.blocksize) != 0) {
 			ploop_err(0, "Image corrupted: delta.l2[%d]=%d",
 					l2_slot, delta.l2[l2_slot]);
 			goto err;
 		}
 		if (delta.l2[l2_slot] != 0) {
-			if (PREAD(&delta, buf, CLUSTER, (delta.l2[l2_slot] << PLOOP1_SECTOR_LOG)))
+			if (PREAD(&delta, buf, cluster, S2B(delta.l2[l2_slot])))
 				goto err;
 		} else {
-			bzero(buf, sizeof(buf));
+			bzero(buf, cluster);
 		}
 
-		if (PWRITE(&odelta, buf, CLUSTER, clu * CLUSTER))
+		if (PWRITE(&odelta, buf, cluster, clu * cluster))
 			goto err;
 	}
 
@@ -1542,6 +1556,7 @@ err:
 	if (ret)
 		unlink(tmp);
 	close_delta(&delta);
+	free(buf);
 
 	return ret;
 }
@@ -1549,41 +1564,25 @@ err:
 static int expanded2preallocated(struct ploop_disk_images_data *di)
 {
 	struct delta delta = {};
-	int i;
-	int off = 16;
 	__u32 clu;
-	void *buf = NULL;
-	int clu_to_fill;
 	off_t data_off;
 	int ret = -1;
+	__u64 cluster;
 
 	ploop_log(0, "Converting image to preallocated...");
+	// FIXME: deny on snapshots
 	if (open_delta(&delta, di->images[0]->file, O_RDWR, OD_OFFLINE))
 		return SYSEXIT_OPEN;
-	if (posix_memalign(&buf, 4096, CLUSTER)) {
-		ploop_err(errno, "posix_memalign");
-		goto err;
-	}
-	bzero(buf, CLUSTER);
-	// First stage: write data blocks
-	// clu_to_fill = total_data_clu - (allocated_clu - total_idx_clu)
-	clu_to_fill = delta.l2_size - (delta.alloc_head - delta.l1_size);
+	cluster = S2B(delta.blocksize);
 	data_off = delta.alloc_head;
-	for (i = 0; i < clu_to_fill; i++) {
-		if (PWRITE(&delta, buf, CLUSTER, (data_off + i) * CLUSTER)) {
-			free(buf);
-			goto err;
-		}
-	}
-	free(buf); buf = NULL;
 	if (fsync(delta.fd)) {
 		ploop_err(errno, "fsync");
 		goto err;
 	}
 	// Second stage: update index
 	for (clu = 0; clu < delta.l2_size; clu++) {
-		int l2_cluster = (clu + off) / (CLUSTER / sizeof(__u32));
-		__u32 l2_slot  = (clu + off) % (CLUSTER / sizeof(__u32));
+		int l2_cluster = (clu + PLOOP_MAP_OFFSET) / (cluster / sizeof(__u32));
+		__u32 l2_slot  = (clu + PLOOP_MAP_OFFSET) % (cluster / sizeof(__u32));
 
 		if (l2_cluster >= delta.l1_size) {
 			ploop_err(0, "abort: l2_cluster >= delta.l1_size");
@@ -1591,13 +1590,19 @@ static int expanded2preallocated(struct ploop_disk_images_data *di)
 		}
 
 		if (delta.l2_cache != l2_cluster) {
-			if (PREAD(&delta, delta.l2, CLUSTER, (off_t)l2_cluster * CLUSTER))
+			if (PREAD(&delta, delta.l2, cluster, (off_t)l2_cluster * cluster))
 				goto err;
 			delta.l2_cache = l2_cluster;
 		}
 		if (delta.l2[l2_slot] == 0) {
-			off_t idx_off = (off_t)l2_cluster * CLUSTER + (l2_slot*sizeof(__u32));
-			delta.l2[l2_slot] = data_off << PLOOP1_DEF_CLUSTER_LOG;
+			off_t idx_off = (off_t)l2_cluster * cluster + (l2_slot*sizeof(__u32));
+
+			delta.l2[l2_slot] = data_off * delta.blocksize;
+
+			if (sys_fallocate(delta.fd, 0, data_off * cluster, cluster)) {
+				ploop_err(errno, "Failed to fallocate");
+				goto err;
+			}
 
 			if (PWRITE(&delta, &delta.l2[l2_slot], sizeof(__u32), idx_off))
 				goto err;
@@ -1645,16 +1650,18 @@ int ploop_get_info(struct ploop_disk_images_data *di, struct ploop_info *info)
 	char mnt[MAXPATHLEN];
 	char dev[64];
 
-	if (di->nimages == 0)
-		return SYSEXIT_PARAM;
+	if (ploop_lock_di(di))
+		return SYSEXIT_LOCK;
 	if (ploop_find_dev_by_uuid(di, 1, dev, sizeof(dev)) == 0 &&
 			get_mount_dir(dev, mnt, sizeof(mnt)) == 0)
 	{
+		ploop_unlock_di(di);
 		if (get_statfs_info(mnt, info))
 			return -1;
 		return 0;
 	}
 
+	ploop_unlock_di(di);
 	return read_statfs_info(di->images[0]->file, info);
 }
 
@@ -1670,7 +1677,7 @@ static int do_snapshot(int lfd, int fd, struct ploop_ctl_delta *req)
 	return 0;
 }
 
-int create_snapshot(const char *device, const char *delta, int syncfs)
+int create_snapshot(const char *device, const char *delta, __u32 blocksize, int syncfs)
 {
 	int ret;
 	int lfd = -1;
@@ -1689,7 +1696,7 @@ int create_snapshot(const char *device, const char *delta, int syncfs)
 		ret = SYSEXIT_BLKDEV;
 		goto err;
 	}
-	bdsize >>= 9;
+	bdsize = bytes2sec(bdsize);
 
 	if (bdsize == 0) {
 		ploop_err(0, "Can't get block device %s size", device);
@@ -1697,7 +1704,7 @@ int create_snapshot(const char *device, const char *delta, int syncfs)
 		goto err;
 	}
 
-	fd = create_empty_delta(delta, bdsize);
+	fd = create_empty_delta(delta, blocksize, bdsize);
 	if (fd < 0) {
 		ret = SYSEXIT_OPEN;
 		goto err;
@@ -1707,7 +1714,7 @@ int create_snapshot(const char *device, const char *delta, int syncfs)
 
 	req.c.pctl_format = PLOOP_FMT_PLOOP1;
 	req.c.pctl_flags = syncfs ? PLOOP_FLAG_FS_SYNC : 0;
-	req.c.pctl_cluster_log = 9;
+	req.c.pctl_cluster_log = ffs(blocksize) - 1;
 	req.c.pctl_size = 0;
 	req.c.pctl_chunks = 1;
 
@@ -1739,6 +1746,10 @@ int ploop_create_snapshot(struct ploop_disk_images_data *di, struct ploop_snapsh
 	char conf_tmp[MAXPATHLEN];
 	int online = 0;
 
+	ret = not_supported_for_vm(di);
+	if (ret)
+		return ret;
+
 	if (di->nimages == 0) {
 		ploop_err(0, "No images");
 		return SYSEXIT_PARAM;
@@ -1786,7 +1797,7 @@ int ploop_create_snapshot(struct ploop_disk_images_data *di, struct ploop_snapsh
 
 	if (ret != 0) {
 		// offline snapshot
-		fd = create_empty_delta(fname, di->size);
+		fd = create_empty_delta(fname, di->blocksize, di->size);
 		if (fd != -1) {
 			ret = 0;
 			close(fd);
@@ -1796,7 +1807,7 @@ int ploop_create_snapshot(struct ploop_disk_images_data *di, struct ploop_snapsh
 	} else {
 		// Always sync fs
 		online = 1;
-		ret = create_snapshot(dev, fname, 1);
+		ret = create_snapshot(dev, fname, di->blocksize, 1);
 	}
 	if (ret)
 		goto err_cleanup2;
@@ -1835,6 +1846,10 @@ int ploop_switch_snapshot(struct ploop_disk_images_data *di, const char *guid, i
 	char conf[MAXPATHLEN];
 	char conf_tmp[MAXPATHLEN];
 
+	ret = not_supported_for_vm(di);
+	if (ret)
+		return ret;
+
 	if (ploop_lock_di(di))
 		return SYSEXIT_LOCK;
 
@@ -1885,7 +1900,7 @@ int ploop_switch_snapshot(struct ploop_disk_images_data *di, const char *guid, i
 		goto err_cleanup1;
 
 	// offline snapshot
-	fd = create_empty_delta(new_top_delta_fnanme, di->size);
+	fd = create_empty_delta(new_top_delta_fnanme, di->blocksize, di->size);
 	if (fd == -1) {
 		ret = SYSEXIT_CREAT;
 		goto err_cleanup2;
@@ -1939,6 +1954,10 @@ int ploop_delete_snapshot(struct ploop_disk_images_data *di, const char *guid)
 	char dev[64];
 	int snap_id;
 
+	ret = not_supported_for_vm(di);
+	if (ret)
+		return ret;
+
 	if (ploop_lock_di(di))
 		return SYSEXIT_LOCK;
 
@@ -1974,7 +1993,7 @@ int ploop_delete_snapshot(struct ploop_disk_images_data *di, const char *guid)
 		ret = ploop_store_diskdescriptor(conf, di);
 		if (ret)
 			goto err;
-		ploop_log(0, "remove %s", fname);
+		ploop_log(0, "Removing %s", fname);
 		if (fname != NULL && unlink(fname)) {
 			ploop_err(errno, "unlink %s", fname);
 			ret = SYSEXIT_UNLINK;
