@@ -35,7 +35,7 @@ static struct ploop_cancel_handle *_s_cancel_handle;
 void usage_summary(void)
 {
 	fprintf(stderr, "Usage: ploop init -s SIZE [-f FORMAT] NEW_DELTA\n"
-			"       ploop mount [-rP] [-b nfs] [-f raw] -d DEVICE [ TOP_DELTA ... ] BASE_DELTA\n"
+			"       ploop mount [-rP] [-f raw] -d DEVICE [ TOP_DELTA ... ] BASE_DELTA\n"
 			"       ploop umount -d DEVICE\n"
 			"       ploop { delete | rm } -d DEVICE -l LEVEL\n"
 			"       ploop merge -d DEVICE [-l LEVEL[..TOP_LEVEL]]\n"
@@ -43,21 +43,22 @@ void usage_summary(void)
 			"       ploop fsck [-fcr] DELTA\n"
 			"       ploop getdev\n"
 			"       ploop resize -s SIZE BASE_DELTA\n"
-			"       ploop snapshot [-F] [-b nfs] -d DEVICE NEW_DELTA\n"
+			"       ploop snapshot [-F] -d DEVICE NEW_DELTA\n"
 			"       ploop snapshot DiskDescriptor.xml\n"
 			"       ploop snapshot-delete -u <uuid> DiskDescriptor.xml\n"
 			"       ploop snapshot-merge [-u <uuid>] DiskDescriptor.xml\n"
 			"       ploop snapshot-switch -u <uuid> DiskDescriptor.xml\n"
-			"Also:  ploop { stat | add | start | stop | clear | replace } ...\n"
+			"Also:  ploop { stat | start | stop | clear } ...\n"
 	       );
 }
 
 static void usage_init(void)
 {
-	fprintf(stderr, "Usage: ploop init -s SIZE [-f FORMAT] [-t FSTYPE] DELTA\n"
+	fprintf(stderr, "Usage: ploop init -s SIZE [-f FORMAT] [-t FSTYPE] [-b BLOCKSIZE] DELTA\n"
 			"       SIZE := NUMBER[kmg], \n"
 			"       FORMAT := { raw | ploop1 }\n"
 			"       DELTA := path to new image file\n"
+			"       BLOCKSIZE := block size\n"
 			"       FSTYPE := make file system\n");
 }
 
@@ -78,7 +79,7 @@ static int plooptool_init(int argc, char **argv)
 	struct ploop_create_param param = {};
 
 	param.mode = PLOOP_EXPANDED_MODE;
-	while ((i = getopt(argc, argv, "s:f:t:")) != EOF) {
+	while ((i = getopt(argc, argv, "s:b:f:t:")) != EOF) {
 		switch (i) {
 		case 's':
 			if (parse_size(optarg, &size_sec)) {
@@ -86,6 +87,16 @@ static int plooptool_init(int argc, char **argv)
 				return -1;
 			}
 			break;
+		case 'b': {
+			  char * endptr;
+
+			  param.blocksize = strtoul(optarg, &endptr, 0);
+			  if (optarg == endptr) {
+				  usage_init();
+				  return -1;
+			  }
+			  break;
+		}
 		case 'f':
 			if (strcmp(optarg, "raw") == 0)
 				param.mode = PLOOP_RAW_MODE;
@@ -132,11 +143,10 @@ static int plooptool_init(int argc, char **argv)
 
 static void usage_mount(void)
 {
-	fprintf(stderr, "Usage: ploop mount [-rP] [-b IO_MODULE] [-f FORMAT] [-d DEVICE] [TOP_DELTA ... ] BASE_DELTA\n"
-			"     : ploop mount [-rP] [-m DIR] [-u UUID] DiskDescriptor.xml\n"
+	fprintf(stderr, "Usage: ploop mount [-rP] [-f FORMAT] [-d DEVICE] [TOP_DELTA ... ] BASE_DELTA\n"
+			"       ploop mount [-rP] [-m DIR] [-u UUID] DiskDescriptor.xml\n"
 			"       FORMAT := { raw | ploop1 }\n"
-			"       IO_MODULE := { direct | nfs }\n"
-			"       DEVICE := ploop device, f.e. /dev/ploop0\n"
+			"       DEVICE := ploop device, e.g. /dev/ploop0\n"
 			"       *DELTA := path to image file\n"
 			"       -r     - mount images read-only\n"
 			"       -P     - rescan partition table upon mount\n"
@@ -152,7 +162,7 @@ static int plooptool_mount(int argc, char **argv)
 	int base = 0;
 	struct ploop_mount_param mountopts = {};
 
-	while ((i = getopt(argc, argv, "rb:f:Pd:m:t:u:o:")) != EOF) {
+	while ((i = getopt(argc, argv, "rf:Pd:m:t:u:o:")) != EOF) {
 		switch (i) {
 		case 'd':
 			strncpy(mountopts.device, optarg, sizeof(mountopts.device)-1);
@@ -168,14 +178,6 @@ static int plooptool_mount(int argc, char **argv)
 				return -1;
 			}
 			break;
-		case 'b':
-			if (strcmp(optarg, "nfs") == 0)
-				mountopts.nfs = 1;
-			else if (strcmp(optarg, "direct") != 0) {
-				usage_mount();
-				return -1;
-			}
-			break;
 		case 'P':
 			mountopts.partitioned = 1;
 			break;
@@ -228,117 +230,16 @@ err:
 		ploop_free_diskdescriptor(di);
 	}
 	else
-		ret = ploop_mount(argv, &mountopts, raw);
+		ret = ploop_mount(NULL, argv, &mountopts, raw);
 
 	return ret;
 
 }
 
-static void usage_add(void)
-{
-	fprintf(stderr, "Usage: ploop add [-w] [-b IO_MODULE] [-f FORMAT] -d DEVICE DELTA\n"
-			"       FORMAT := { raw | ploop1 }\n"
-			"       IO_MODULE := { direct | nfs }\n"
-			"       DEVICE := ploop device, f.e. /dev/ploop0\n"
-			"       DELTA := path to image file\n");
-}
-
-static int plooptool_add(int argc, char **argv)
-{
-	int i;
-	int lfd;
-	int fd;
-	struct {
-		struct ploop_ctl c;
-		struct ploop_ctl_chunk f;
-	} req;
-
-	struct {
-		int rw;
-		int raw;
-		int nfs;
-		char * device;
-	} addopts = { };
-
-	while ((i = getopt(argc, argv, "wb:f:d:")) != EOF) {
-		switch (i) {
-		case 'd':
-			addopts.device = optarg;
-			break;
-		case 'w':
-			addopts.rw = 1;
-			break;
-		case 'f':
-			if (strcmp(optarg, "raw") == 0)
-				addopts.raw = 1;
-			else if (strcmp(optarg, "ploop1") != 0) {
-				usage_add();
-				return -1;
-			}
-			break;
-		case 'b':
-			if (strcmp(optarg, "nfs") == 0)
-				addopts.nfs = 1;
-			else if (strcmp(optarg, "direct") != 0) {
-				usage_add();
-				return -1;
-			}
-			break;
-		default:
-			usage_add();
-			return -1;
-		}
-	}
-
-	argc -= optind;
-	argv += optind;
-
-	if (argc < 1 || !addopts.device) {
-		usage_add();
-		return -1;
-	}
-
-	lfd = open(addopts.device, O_RDONLY);
-	if (lfd < 0) {
-		perror("Cannot open device");
-		return SYSEXIT_DEVICE;
-	}
-
-
-	fd = open(argv[0], addopts.rw ? O_RDWR : O_RDONLY);
-	if (fd < 0) {
-		perror("open file");
-		return SYSEXIT_OPEN;
-	}
-
-	memset(&req, 0, sizeof(req));
-
-	req.c.pctl_format = PLOOP_FMT_PLOOP1;
-	if (addopts.raw)
-		req.c.pctl_format = PLOOP_FMT_RAW;
-	if (!addopts.rw)
-		req.c.pctl_flags = PLOOP_FMT_RDONLY;
-	req.c.pctl_cluster_log = 9;
-	req.c.pctl_size = 0;
-	req.c.pctl_chunks = 1;
-
-	req.f.pctl_fd = fd;
-	req.f.pctl_type = PLOOP_IO_DIRECT;
-	if (addopts.nfs)
-		req.f.pctl_type = PLOOP_IO_NFS;
-
-	if (ioctl(lfd, PLOOP_IOC_ADD_DELTA, &req) < 0) {
-		perror("PLOOP_IOC_ADD_DELTA");
-		return SYSEXIT_DEVIOC;
-	}
-
-	return 0;
-}
-
 static void usage_start(void)
 {
-	fprintf(stderr, "Usage: ploop start [-P] -d DEVICE\n" 
-			"       DEVICE := ploop device, f.e. /dev/ploop0\n"
+	fprintf(stderr, "Usage: ploop start [-P] -d DEVICE\n"
+			"       DEVICE := ploop device, e.g. /dev/ploop0\n"
 			"       -P     - rescan partition table\n");
 }
 
@@ -375,7 +276,7 @@ static int plooptool_start(int argc, char **argv)
 
 	lfd = open(startopts.device, O_RDONLY);
 	if (lfd < 0) {
-		perror("Cannot open device");
+		perror("Can't open device");
 		return SYSEXIT_DEVICE;
 	}
 
@@ -394,8 +295,8 @@ static int plooptool_start(int argc, char **argv)
 
 static void usage_stop(void)
 {
-	fprintf(stderr, "Usage: ploop stop -d DEVICE\n" 
-			"       DEVICE := ploop device, f.e. /dev/ploop0\n");
+	fprintf(stderr, "Usage: ploop stop -d DEVICE\n"
+			"       DEVICE := ploop device, e.g. /dev/ploop0\n");
 }
 
 static int plooptool_stop(int argc, char **argv)
@@ -427,7 +328,7 @@ static int plooptool_stop(int argc, char **argv)
 
 	lfd = open(stopopts.device, O_RDONLY);
 	if (lfd < 0) {
-		perror("Cannot open device");
+		perror("Can't open device");
 		return SYSEXIT_DEVICE;
 	}
 
@@ -441,8 +342,8 @@ static int plooptool_stop(int argc, char **argv)
 
 static void usage_clear(void)
 {
-	fprintf(stderr, "Usage: ploop clear -d DEVICE\n" 
-			"       DEVICE := ploop device, f.e. /dev/ploop0\n");
+	fprintf(stderr, "Usage: ploop clear -d DEVICE\n"
+			"       DEVICE := ploop device, e.g. /dev/ploop0\n");
 }
 
 static int plooptool_clear(int argc, char **argv)
@@ -474,7 +375,7 @@ static int plooptool_clear(int argc, char **argv)
 
 	lfd = open(stopopts.device, O_RDONLY);
 	if (lfd < 0) {
-		perror("Cannot open device");
+		perror("Can't open device");
 		return SYSEXIT_DEVICE;
 	}
 
@@ -489,7 +390,7 @@ static int plooptool_clear(int argc, char **argv)
 static void usage_umount(void)
 {
 	fprintf(stderr, "Usage: ploop umount -d DEVICE\n"
-			"       DEVICE := ploop device, f.e. /dev/ploop0\n");
+			"       DEVICE := ploop device, e.g. /dev/ploop0\n");
 }
 
 static int plooptool_umount(int argc, char **argv)
@@ -552,7 +453,7 @@ static void usage_rm(void)
 {
 	fprintf(stderr, "Usage: ploop { delete | rm } -d DEVICE -l LEVEL\n"
 			"       LEVEL := NUMBER, distance from base delta\n"
-			"       DEVICE := ploop device, f.e. /dev/ploop0\n");
+			"       DEVICE := ploop device, e.g. /dev/ploop0\n");
 }
 
 static int plooptool_rm(int argc, char **argv)
@@ -602,114 +503,11 @@ static int plooptool_rm(int argc, char **argv)
 	return 0;
 }
 
-static void usage_replace(void)
-{
-	fprintf(stderr, "Usage: ploop replace [-b IO_MODULE] [-f FORMAT] -l LEVEL -d DEVICE DELTA\n"
-			"       LEVEL := NUMBER, distance from base\n"
-			"       FORMAT := { raw | ploop1 }\n"
-			"       IO_MODULE := { direct | nfs }\n"
-			"       DEVICE := ploop device, f.e. /dev/ploop0\n"
-			"       DELTA := path to new image file\n");
-}
-
-static int plooptool_replace(int argc, char **argv)
-{
-	int i;
-	int lfd, fd;
-	struct {
-		struct ploop_ctl c;
-		struct ploop_ctl_chunk f;
-	} req;
-
-	struct {
-		char * device;
-		char * deltafile;
-		int level;
-		int nfs;
-		int raw;
-	} replaceopts = { .level = -1, };
-
-	while ((i = getopt(argc, argv, "b:f:l:d:")) != EOF) {
-		switch (i) {
-		case 'b':
-			if (strcmp(optarg, "nfs") == 0)
-				replaceopts.nfs = 1;
-			else if (strcmp(optarg, "direct") != 0) {
-				usage_replace();
-				return -1;
-			}
-			break;
-		case 'f':
-			if (strcmp(optarg, "raw") == 0)
-				replaceopts.raw = 1;
-			else if (strcmp(optarg, "direct") != 0) {
-				usage_replace();
-				return -1;
-			}
-			break;
-		case 'l':
-			replaceopts.level = atoi(optarg);
-			break;
-		case 'd':
-			replaceopts.device = optarg;
-			break;
-		default:
-			usage_replace();
-			return -1;
-		}
-	}
-
-	argc -= optind;
-	argv += optind;
-
-	if (argc != 1 || !replaceopts.device ||
-	    replaceopts.level < 0 || replaceopts.level > 127) {
-		usage_replace();
-		return -1;
-	}
-
-	replaceopts.deltafile = argv[0];
-
-	lfd = open(replaceopts.device, O_RDONLY);
-	if (lfd < 0) {
-		perror("Cannot open device");
-		return SYSEXIT_DEVICE;
-	}
-
-	fd = open(replaceopts.deltafile, O_RDONLY);
-	if (fd < 0) {
-		perror("open file");
-		return SYSEXIT_OPEN;
-	}
-
-	memset(&req, 0, sizeof(req));
-
-	req.c.pctl_format = PLOOP_FMT_PLOOP1;
-	if (replaceopts.raw)
-		req.c.pctl_format = PLOOP_FMT_RAW;
-	req.c.pctl_flags = PLOOP_FMT_RDONLY;
-	req.c.pctl_cluster_log = 9;
-	req.c.pctl_size = 0;
-	req.c.pctl_chunks = 1;
-	req.c.pctl_level = replaceopts.level;
-
-	req.f.pctl_fd = fd;
-	req.f.pctl_type = PLOOP_IO_DIRECT;
-	if (replaceopts.nfs)
-		req.f.pctl_type = PLOOP_IO_NFS;
-
-	if (ioctl(lfd, PLOOP_IOC_REPLACE_DELTA, &req) < 0) {
-		perror("PLOOP_IOC_REPLACE_DELTA");
-		return SYSEXIT_DEVIOC;
-	}
-	return 0;
-}
-
 static void usage_snapshot(void)
 {
 	fprintf(stderr, "Usage: ploop snapshot [-u <uuid>] DiskDescriptor.xml\n"
 			"       ploop snapshot [-F] -d DEVICE DELTA\n"
-			"         DEVICE := ploop device, f.e. /dev/ploop0\n"
+			"         DEVICE := ploop device, e.g. /dev/ploop0\n"
 			"         DELTA := path to new image file\n"
 			"         -F    - synchronize file system before taking snapshot\n"
 		);
@@ -757,11 +555,14 @@ static int plooptool_snapshot(int argc, char **argv)
 
 		ploop_free_diskdescriptor(di);
 	} else {
+		__u32 blocksize = 0;
 		if (!device) {
 			usage_snapshot();
 			return -1;
 		}
-		ret = create_snapshot(device, argv[0], syncfs);
+		if (ploop_get_attr(device, "block_size", (int*) &blocksize))
+			return 1;
+		ret = create_snapshot(device, argv[0], blocksize, syncfs);
 	}
 
 	return ret;
@@ -913,32 +714,22 @@ static void usage_getdevice(void)
 
 static int plooptool_getdevice(int argc, char **argv)
 {
-	int lfd;
-	struct ploop_getdevice_ctl req;
+	int minor, ret;
 
 	if (argc != 1) {
 		usage_getdevice();
 		return -1;
 	}
+	ret = ploop_getdevice(&minor);
+	if (ret == 0)
+		printf("Next unused minor: %d\n", minor);
 
-	lfd = open("/dev/ploop0", O_RDONLY);
-	if (lfd < 0) {
-		perror("Cannot open device");
-		return SYSEXIT_DEVICE;
-	}
-
-	if (ioctl(lfd, PLOOP_IOC_GETDEVICE, &req) < 0) {
-		perror("PLOOP_IOC_GETDEVICE");
-		return SYSEXIT_DEVIOC;
-	}
-
-	printf("Next unused minor: %d\n", req.minor);
-	return 0;
+	return ret;
 }
 
 static void usage_resize(void)
 {
-	fprintf(stderr, "Usage: resize -s NEW_SIZE DiskDescriptor.xml\n");
+	fprintf(stderr, "Usage: ploop resize -s NEW_SIZE DiskDescriptor.xml\n");
 }
 
 static int plooptool_resize(int argc, char **argv)
@@ -990,7 +781,7 @@ static int plooptool_resize(int argc, char **argv)
 
 static void usage_convert(void)
 {
-	fprintf(stderr, "Usage: convert -t <raw|preallocated>\n");
+	fprintf(stderr, "Usage: ploop convert -t <raw|preallocated>\n");
 }
 
 static int plooptool_convert(int argc, char **argv)
@@ -1039,7 +830,7 @@ static int plooptool_convert(int argc, char **argv)
 
 static void usage_info(void)
 {
-	fprintf(stderr, "Usage: info\n");
+	fprintf(stderr, "Usage: ploop info\n");
 }
 
 static void print_info(struct ploop_info *info)
@@ -1118,8 +909,6 @@ int main(int argc, char **argv)
 
 	if (strcmp(cmd, "init") == 0)
 		return plooptool_init(argc, argv);
-	if (strcmp(cmd, "add") == 0)
-		return plooptool_add(argc, argv);
 	if (strcmp(cmd, "start") == 0)
 		return plooptool_start(argc, argv);
 	if (strcmp(cmd, "stop") == 0)
@@ -1132,8 +921,6 @@ int main(int argc, char **argv)
 		return plooptool_umount(argc, argv);
 	if (strcmp(cmd, "delete") == 0 || strcmp(cmd, "rm") == 0)
 		return plooptool_rm(argc, argv);
-	if (strcmp(cmd, "replace") == 0)
-		return plooptool_replace(argc, argv);
 	if (strcmp(cmd, "snapshot") == 0)
 		return plooptool_snapshot(argc, argv);
 	if (strcmp(cmd, "snapshot-switch") == 0)
